import pytest

from youtube_wpm._youtube import normalize_youtube_id


class Test_print_normalize_youtube_id:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            ["QpBTM0GO6xI", "QpBTM0GO6xI"],
            ["https://www.youtube.com/watch?v=QpBTM0GO6xI", "QpBTM0GO6xI"],
            ["https://www.youtube.com/watch?v=QpBTM0GO6xI?t=14", "QpBTM0GO6xI"],
            ["https://youtu.be/QpBTM0GO6xI", "QpBTM0GO6xI"],
            ["https://youtu.be/QpBTM0GO6xI?t=14", "QpBTM0GO6xI"],
            ["https://youtu.be/QpBTM0GO6xI?si=ABCDEFGHIJKLMNOP&t=24", "QpBTM0GO6xI"],
            ["https://www.youtube.com/watch?v=QpBTM0GO6xI&feature=youtu.be", "QpBTM0GO6xI"],
        ],
    )
    def test_normal(self, value, expected):
        assert normalize_youtube_id(value) == expected

    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            ["INVALID_VIDEO_ID", ValueError],
            ["https://www.youtube.com/watch?v=INVALID", ValueError],
            ["https://youtu.be/INVALID?t=14", ValueError],
        ],
    )
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            normalize_youtube_id(value)



import argparse
import sys
from decimal import Decimal
from textwrap import dedent

import humanreadable as hr
from youtube_transcript_api import NoTranscriptFound, TranscriptsDisabled, YouTubeTranscriptApi

from .__version__ import __version__
from ._const import EXIT_WPM_DIFF_THRESHOLD, INITIAL_APPROXIMATE_WPM, MAX_ITERATION, MODULE_NAME
from ._logger import LogLevel, initialize_logger, logger
from ._youtube import calc_speak_time, normalize_youtube_id


def parse_option() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=dedent(
            """\
            A CLI tool to get a YouTube video's words per minute (WPM).
            """
        ),
        epilog=dedent(
            f"""\
            Issue tracker: https://github.com/thombashi/{MODULE_NAME}/issues
            """
        ),
    )
    parser.add_argument("video_id", help="Youtube video ID")

    parser.add_argument("-V", "--version", action="version", version=f"%(prog)s {__version__}")

    parser.add_argument(
        "--language",
        default="en",
        help="language code of the transcript. defaults to 'en'.",
    )
    parser.add_argument(
        "--initial-wpm",
        type=float,
        default=INITIAL_APPROXIMATE_WPM,
        help=f"initial approximate words per minute. defaults to {INITIAL_APPROXIMATE_WPM}.",
    )

    loglevel_dest = "log_level"
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--debug",
        dest=loglevel_dest,
        action="store_const",
        const=LogLevel.DEBUG,
        default=LogLevel.INFO,
        help="for debug print.",
    )
    group.add_argument(
        "--quiet",
        dest=loglevel_dest,
        action="store_const",
        const=LogLevel.QUIET,
        default=LogLevel.INFO,
        help="suppress execution log messages.",
    )

    return parser.parse_args()


def calc_seconds_per_word(wpm: Decimal) -> Decimal:
    return 60 / wpm


def main() -> int:
    ns = parse_option()

    initialize_logger(name=MODULE_NAME, log_level=ns.log_level)

    video_id = normalize_youtube_id(ns.video_id)
    logger.debug(f"fetching transcripts of {video_id} ...")

    try:
        transcripts = YouTubeTranscriptApi.list_transcripts(video_id)
    except TranscriptsDisabled as e:
        logger.error(e)
        return 1

    try:
        transcript = transcripts.find_transcript([ns.language])
    except NoTranscriptFound as e:
        logger.error(e)
        return 1

    logger.debug(
        f"{video_id=}, {transcript.language=}, {transcript.language_code=}, {transcript.is_generated=}"
    )

    initial_wpm: Decimal = Decimal(200)
    prev_wpm: Decimal = initial_wpm
    inference_spw = calc_seconds_per_word(prev_wpm)

    sequences = transcript.fetch()
    for i in range(MAX_ITERATION):
        stat = calc_speak_time(sequences, inference_spw=inference_spw)
        diff_wpm = abs(stat.wpm - prev_wpm)
        logger.debug(f"iteration={i}, wpm={stat.wpm:.1f}, diff_wpm={diff_wpm:.1f}")
        if diff_wpm < EXIT_WPM_DIFF_THRESHOLD:
            break

        inference_spw = calc_seconds_per_word(stat.wpm)
        prev_wpm = stat.wpm

    assert stat.total_speak_secs > 0
    outputs = [
        f"Total word count: {stat.total_word_ct}",
        "Approximate blank time: {}".format(
            hr.Time(
                str(int(stat.total_blank_secs)), default_unit=hr.Time.Unit.SECOND
            ).to_humanreadable()
        ),
        "Approximate speaking time: {}".format(
            hr.Time(
                str(int(stat.total_speak_secs)), default_unit=hr.Time.Unit.SECOND
            ).to_humanreadable()
        ),
        f"Words Per Minute: {stat.wpm:.1f}",
    ]

    for out in outputs:
        print(out)

    return 0


if __name__ == "__main__":
    sys.exit(main())



from typing import Final


__author__: Final[str] = "Tsuyoshi Hombashi"
__copyright__: Final[str] = f"Copyright 2023, {__author__}"
__license__: Final[str] = "MIT License"
__version__: Final[str] = "0.0.1"
__maintainer__: Final[str] = __author__
__email__: Final[str] = "tsuyoshi.hombashi@gmail.com"



import re
from dataclasses import dataclass
from decimal import Decimal
from typing import Dict, Final, List, Pattern

from ._logger import logger


ID_PATTERN: Final[str] = r"[a-zA-Z0-9_-]{11}"
RE_VIDEO_ID: Final[Pattern] = re.compile(rf"^({ID_PATTERN})$")
RE_LONG_VIDEO_URL: Final[Pattern] = re.compile(
    rf"(?<=https://www\.youtube.com/watch\?v=)({ID_PATTERN})"
)
RE_SHORT_VIDEO_URL: Final[Pattern] = re.compile(rf"(?<=https://youtu\.be/)({ID_PATTERN})")
RE_SOUND: Final[Pattern] = re.compile(r"^\[[A-Z][a-zA-Z ]+\]$")


@dataclass
class SpeakStat:
    total_word_ct: int
    total_blank_secs: Decimal
    total_speak_secs: Decimal

    @property
    def wpm(self) -> Decimal:
        return self.total_word_ct / self.total_speak_secs * 60


@dataclass(frozen=True)
class Weight:
    display_duration: Decimal
    approx_speak_time: Decimal

    @property
    def _total_wegiht(self) -> Decimal:
        return self.display_duration + self.approx_speak_time

    def calc_display_duration_weight(self) -> Decimal:
        return self.display_duration / self._total_wegiht

    def calc_approx_speak_time_weight(self) -> Decimal:
        return self.approx_speak_time / self._total_wegiht


def validate_youtube_id(value: str) -> bool:
    return RE_VIDEO_ID.match(value) is not None


def normalize_youtube_id(value: str) -> str:
    if validate_youtube_id(value):
        return value

    m = RE_LONG_VIDEO_URL.search(value)
    if m is not None:
        return m.group(1)

    m = RE_SHORT_VIDEO_URL.search(value)
    if m is not None:
        return m.group(1)

    raise ValueError(f"invalid video id: {value}")


def calc_speak_time(sequences: List[Dict], inference_spw: Decimal) -> SpeakStat:
    stat = SpeakStat(
        total_word_ct=0,
        total_blank_secs=Decimal(0),
        total_speak_secs=Decimal(0),
    )
    weight = Weight(
        display_duration=Decimal(1),
        approx_speak_time=Decimal(5),
    )
    begin_talk: Decimal = Decimal(0)
    last: Decimal = Decimal(0)
    prev_last: Decimal = Decimal(0)

    for sequence in sequences:
        text = sequence["text"].strip()
        if RE_SOUND.search(text):
            continue

        prev_last = last
        start: Decimal = Decimal(sequence["start"])

        if begin_talk == 0:
            begin_talk = start
        elif prev_last > 0:
            expected_blank_time = start - prev_last
            if expected_blank_time > 0:
                stat.total_blank_secs += start - prev_last

        # duration is how long it is displayed. As there often is some overlap between subtitle boxes
        # https://github.com/jdepoix/youtube-transcript-api/issues/21
        display_duration = Decimal(sequence["duration"])

        words = text.strip().split()
        word_ct = len(words)
        char_ct = sum([len(word) for word in words])
        words_per_duration = word_ct / display_duration
        chars_per_duration = char_ct / display_duration
        approx_speak_time = (word_ct * Decimal(0.95) + char_ct / Decimal(100)) * inference_spw
        logger.trace(
            f"{text}: {start=}, {word_ct=}, {char_ct=}, {display_duration=}, "
            f"words/d={words_per_duration:.2f}, chars/d={chars_per_duration:.2f}, "
            f"speak-time={approx_speak_time:.2f}"
        )

        last = start + (
            display_duration * weight.calc_display_duration_weight()
            + approx_speak_time * weight.calc_approx_speak_time_weight()
        )

        stat.total_word_ct += word_ct

    stat.total_speak_secs = last - begin_talk - stat.total_blank_secs

    return stat
